import os

MORIARTY_BRIDGE_NAME_ENV = "MORIARTY_BRIDGE_NAME"
MORIARTY_SPAWNER_NAME_ENV = "MORIARTY_SPAWNER_NAME"
MORIARTY_MATRIX_TOKEN_ENV = "MORIARTY_MATRIX_TOKEN"
MORIARTY_CONNECTOR_ENDPOINT_ENV = "MORIARTY_CONNECTOR_ENDPOINT"
MORIARTY_BRIDGE_RESULT_QUEUE_URL_ENV = "MORIARTY_BRIDGE_RESULT_QUEUE_URL"

MORIARTY_KUBE_NAMESPACE_ENV = "MORIARTY_KUBE_NAMESPACE"


def get_bridge_name() -> str:
    return os.getenv(MORIARTY_BRIDGE_NAME_ENV, "sqs")


def get_bridge_result_queue_url() -> str:
    return os.getenv(MORIARTY_BRIDGE_RESULT_QUEUE_URL_ENV, "")


def get_spawner_name() -> str:
    return os.getenv(MORIARTY_SPAWNER_NAME_ENV, "kube")


DB_HOST_ENV = "DB_HOST"
DB_PORT_ENV = "DB_PORT"
DB_USER_ENV = "DB_USER"
DB_PASSWORD_ENV = "DB_PASSWORD"
DB_DATABASE_ENV = "DB_DATABASE"


REDIS_URL_ENV = "REDIS_URL"
REDIS_HOST_ENV = "REDIS_HOST"
REDIS_PORT_ENV = "REDIS_PORT"
REDIS_DB_ENV = "REDIS_DB"
REDIS_CLUSTER_ENV = "REDIS_CLUSTER"
REDIS_TLS_ENV = "REDIS_TLS"
REDIS_USERNAME_ENV = "REDIS_USERNAME"
REDIS_PASSWORD_ENV = "REDIS_PASSWORD"


ENDPOINT_NAME_ENV = "ENDPOINT_NAME"
INVOKE_HOST_ENV = "INVOKE_HOST"
INVOKE_PORT_ENV = "INVOKE_PORT"
INVOKE_PATH_ENV = "INVOKE_PATH"
HEALTH_CHECK_PATH_ENV = "HEALTH_CHECK_PATH"
ENABLE_RETRY_ENV = "ENABLE_RETRY"
ENABLE_SSL_ENV = "ENABLE_SSL"
CONCURRENCY_ENV = "CONCURRENCY"

PROCESS_TIMEOUT_ENV = "PROCESS_TIMEOUT"
HEALTH_CHECK_TIMEOUT_ENV = "HEALTH_CHECK_TIMEOUT"
HEALTH_CHECK_INTERVAL_ENV = "HEALTH_CHECK_INTERVAL"


MODEL_S3_ACCESS_KEY_ID_ENV = "MODEL_S3_ACCESS_KEY_ID"
MODEL_S3_SECRET_ACCESS_KEY_ENV = "MODEL_S3_SECRET_ACCESS_KEY"
MODEL_S3_ENDPOINT_URL_ENV = "MODEL_S3_ENDPOINT_URL"
MODEL_AWS_REGION_ENV = "MODEL_AWS_REGION"

MORIARTY_MATRIX_API_URL_ENV = "MORIARTY_MATRIX_API_URL"
